from typing import List as TList

import discord
from cpl_core.database.context import DatabaseContextABC
from cpl_discord.command import DiscordCommandABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_translation import TranslatePipe
from discord import app_commands
from discord.ext import commands
from discord.ext.commands import Context

from bot_core.abc.client_utils_service_abc import ClientUtilsServiceABC
from bot_core.abc.message_service_abc import MessageServiceABC
from bot_core.logging.command_logger import CommandLogger
from bot_data.abc.level_repository_abc import LevelRepositoryABC
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.model.level import Level
from modules.level.level_seeder import LevelSeeder
from modules.level.service.level_service import LevelService
from modules.permission.abc.permission_service_abc import PermissionServiceABC


class LevelGroup(DiscordCommandABC):

    def __init__(
            self,
            logger: CommandLogger,
            message_service: MessageServiceABC,
            bot: DiscordBotServiceABC,
            client_utils: ClientUtilsServiceABC,
            permission_service: PermissionServiceABC,
            translate: TranslatePipe,
            db: DatabaseContextABC,
            levels: LevelRepositoryABC,
            servers: ServerRepositoryABC,
            users: UserRepositoryABC,
            level_service: LevelService,
            level_seeder: LevelSeeder,
    ):
        DiscordCommandABC.__init__(self)

        self._logger = logger
        self._message_service = message_service
        self._bot = bot
        self._client_utils = client_utils
        self._permissions = permission_service
        self._t = translate
        self._db = db
        self._levels = levels
        self._servers = servers
        self._users = users

        self._level_service = level_service
        self._level_seeder = level_seeder

        self._logger.trace(__name__, f'Loaded command service: {type(self).__name__}')

    @commands.hybrid_group()
    @commands.guild_only()
    async def level(self, ctx: Context):
        pass

    @level.command(alias='levels')
    @commands.guild_only()
    async def list(self, ctx: Context, wait: int = None):
        self._logger.debug(__name__, f'Received command level list {ctx}')
        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if not self._permissions.is_member_moderator(ctx.author):
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            self._logger.trace(__name__, f'Finished command level list')
            return

        if ctx.guild is None:
            return

        server = self._servers.get_server_by_discord_id(ctx.guild.id)
        levels = self._levels.get_levels_by_server_id(server.server_id)
        if levels.count() < 1:
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.error.nothing_found'))
            self._logger.trace(__name__, f'Finished command level list')
            return

        level_name = ''
        xp = ''
        permissions = ''
        for level in levels:
            level_name += f'\n{level.name}'
            xp += f'\n{level.min_xp}'
            permissions += f'\n{level.permissions}'

        embed = discord.Embed(
            title=self._t.transform('modules.level.list.title'),
            description=self._t.transform('modules.level.list.description'),
            color=int('ef9d0d', 16)
        )
        embed.add_field(name=self._t.transform('modules.level.list.name'), value=level_name, inline=True)
        embed.add_field(name=self._t.transform('modules.level.list.min_xp'), value=xp, inline=True)
        embed.add_field(name=self._t.transform('modules.level.list.permission_int'), value=permissions, inline=True)
        await self._message_service.send_ctx_msg(ctx, embed, wait_before_delete=wait)
        self._logger.trace(__name__, f'Finished command level list')

    @level.command()
    @commands.guild_only()
    async def create(self, ctx: Context, name: str, color: str, min_xp: int, permissions: int):
        self._logger.debug(__name__, f'Received command level create {ctx}')

        try:
            color = hex(discord.Colour.from_str(color).value)
        except Exception as e:
            self._logger.error(__name__, f'Error parsing color {color}', e)
            return

        try:
            permissions = discord.Permissions(permissions).value
        except Exception as e:
            self._logger.error(__name__, f'Error parsing permissions {permissions}', e)
            return

        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if ctx.guild is None:
            return

        server = self._servers.get_server_by_discord_id(ctx.guild.id)
        level = Level(name, color, min_xp, permissions, server)
        levels = self._levels.get_levels_by_server_id(server.server_id)

        if levels.where(lambda l: l.name == level.name).first_or_default() is not None:
            self._logger.debug(__name__, f'Level with name {level.name} already exists')
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.error.level_with_name_already_exists').format(level.name))
        elif levels.where(lambda l: l.min_xp == level.min_xp).first_or_default() is not None:
            self._logger.debug(__name__, f'Level with min_xp {level.min_xp} already exists {level.name}')
            found_level = levels.where(lambda l: l.min_xp == level.min_xp).first_or_default()
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.error.level_with_xp_already_exists').format(found_level.name, found_level.min_xp))
        else:
            try:
                self._levels.add_level(level)
                self._db.save_changes()
                self._logger.info(__name__, f'Saved level {name} with color {color}, min_xp {min_xp} and permissions {permissions}')
            except Exception as e:
                self._logger.error(__name__, f'Could not save level {name} with color {color}, min_xp {min_xp} and permissions {permissions}', e)
            else:
                await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.create.created').format(name, permissions))

            # send message to ctx.channel because send_ctx_msg resolves ctx
            try:
                await self._message_service.send_channel_message(ctx.channel, self._t.transform('modules.level.create.seeding_started'))
                await self._level_seeder.seed()
                await self._message_service.send_channel_message(ctx.channel, self._t.transform('modules.level.create.seeding_finished'))
            except Exception as e:
                self._logger.error(__name__, f'Level seeding failed', e)
                await self._message_service.send_channel_message(ctx.channel, self._t.transform('modules.level.create.seeding_failed'))

        self._logger.trace(__name__, f'Finished command level create')

    @create.autocomplete('color')
    async def create_color_autocomplete(self, interaction: discord.Interaction, current: str) -> TList[app_commands.Choice[str]]:
        colors = [
            ('blue', discord.Colour.blue().to_rgb()),
            ('dark_blue', discord.Colour.dark_blue().to_rgb()),
            ('dark_gold', discord.Colour.dark_gold().to_rgb()),
            ('dark_gray', discord.Colour.dark_gray().to_rgb()),
            ('dark_green', discord.Colour.dark_green().to_rgb()),
            ('dark_grey', discord.Colour.dark_grey().to_rgb()),
            ('dark_magenta', discord.Colour.dark_magenta().to_rgb()),
            ('dark_orange', discord.Colour.dark_orange().to_rgb()),
            ('dark_purple', discord.Colour.dark_purple().to_rgb()),
            ('dark_red', discord.Colour.dark_red().to_rgb()),
            ('dark_teal', discord.Colour.dark_teal().to_rgb()),
            ('default', discord.Colour.default().to_rgb()),
            ('gold', discord.Colour.gold().to_rgb()),
            ('green', discord.Colour.green().to_rgb()),
            ('greyple', discord.Colour.greyple().to_rgb()),
            ('light_grey', discord.Colour.light_grey().to_rgb()),
            ('magenta', discord.Colour.magenta().to_rgb()),
            ('orange', discord.Colour.orange().to_rgb()),
            ('purple', discord.Colour.purple().to_rgb()),
            ('red', discord.Colour.red().to_rgb()),
            ('teal', discord.Colour.teal().to_rgb()),
            ('yellow', discord.Colour.yellow().to_rgb())
        ]
        # value in rg format see:
        # https://discordpy.readthedocs.io/en/latest/api.html#discord.Colour.to_rgb
        return [app_commands.Choice(name=self._t.transform(f'common.colors.{color}'), value=f'rgb({code[0]}, {code[1]}, {code[2]})') for color, code in colors]

    @level.command()
    @commands.guild_only()
    async def down(self, ctx: Context, member: discord.Member):
        self._logger.debug(__name__, f'Received command level down {ctx} {member}')
        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if not self._permissions.is_member_moderator(ctx.author):
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            self._logger.trace(__name__, f'Finished command level down')
            return

        if ctx.guild is None:
            return

        if member.bot:
            return

        server = self._servers.get_server_by_discord_id(ctx.guild.id)
        user = self._users.get_user_by_discord_id_and_server_id(member.id, server.server_id)
        level = self._level_service.get_level(user)
        levels = self._levels.get_levels_by_server_id(server.server_id).order_by(lambda l: l.min_xp)

        if level == levels.first():
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.down.already_first').format(member.name))
            self._logger.trace(__name__, f'Finished command level down')
            return

        try:
            new_level = levels.where(lambda l: l.min_xp < level.min_xp).last()
            user.xp = new_level.min_xp
            self._users.update_user(user)
            self._db.save_changes()
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.down.success').format(member.name, new_level.name))
            await self._level_service.set_level(user)
        except Exception as e:
            self._logger.error(__name__, f'Cannot level down {member.name} with level {level.name}', e)
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.down.failed').format(member.name))

        self._logger.trace(__name__, f'Finished command level down')

    @level.command()
    @commands.guild_only()
    async def up(self, ctx: Context, member: discord.Member):
        self._logger.debug(__name__, f'Received command level up {ctx} {member}')
        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if not self._permissions.is_member_moderator(ctx.author):
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            self._logger.trace(__name__, f'Finished command level up')
            return

        if ctx.guild is None:
            return

        if member.bot:
            return

        server = self._servers.get_server_by_discord_id(ctx.guild.id)
        user = self._users.get_user_by_discord_id_and_server_id(member.id, server.server_id)
        level = self._level_service.get_level(user)
        levels = self._levels.get_levels_by_server_id(server.server_id).order_by(lambda l: l.min_xp)

        if level.name == levels.last().name:
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.up.already_last').format(member.name))
            self._logger.trace(__name__, f'Finished command level up')
            return

        try:
            new_level = levels.where(lambda l: l.min_xp > level.min_xp).first()
            user.xp = new_level.min_xp
            self._users.update_user(user)
            self._db.save_changes()
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.up.success').format(member.name, new_level.name))
            await self._level_service.set_level(user)
        except Exception as e:
            self._logger.error(__name__, f'Cannot level up {member.name} with level {level.name}', e)
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.up.failed').format(member.name))

        self._logger.trace(__name__, f'Finished command level up')

    @level.command()
    @commands.guild_only()
    async def set(self, ctx: Context, member: discord.Member, level: str):
        self._logger.debug(__name__, f'Received command level up {ctx} {member}')
        if not await self._client_utils.check_if_bot_is_ready_yet_and_respond(ctx):
            return

        if not self._permissions.is_member_moderator(ctx.author):
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            self._logger.trace(__name__, f'Finished command level up')
            return

        if ctx.guild is None:
            return

        if member.bot:
            return

        server = self._servers.get_server_by_discord_id(ctx.guild.id)
        user = self._users.get_user_by_discord_id_and_server_id(member.id, server.server_id)
        current_level = self._level_service.get_level(user)
        new_level = self._levels.get_levels_by_server_id(server.server_id).where(lambda l: l.name == level).single_or_default()

        if new_level is None:
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.set.failed').format(member.name, level))
            self._logger.trace(__name__, f'Finished command level set')
            return

        if current_level.name == level:
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.set.already_level').format(member.name, level))
            self._logger.trace(__name__, f'Finished command level set')
            return

        try:
            user.xp = new_level.min_xp
            self._users.update_user(user)
            self._db.save_changes()
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.set.success').format(member.name, new_level.name))
            await self._level_service.set_level(user)
        except Exception as e:
            self._logger.error(__name__, f'Cannot set level {level} for {member.name}', e)
            await self._message_service.send_ctx_msg(ctx, self._t.transform('modules.level.set.failed').format(member.name))

        self._logger.trace(__name__, f'Finished command level set')

    @set.autocomplete('level')
    async def set_autocomplete(self, interaction: discord.Interaction, current: str) -> TList[app_commands.Choice[str]]:
        server = self._servers.get_server_by_discord_id(interaction.guild.id)
        levels = self._levels.get_levels_by_server_id(server.server_id).select(lambda l: l.name)
        return [app_commands.Choice(name=level, value=level) for level in levels]
